#include "TascamUs1641.h"

using namespace std;

/**
 * Initialize variables and USB communication.
 * Initialize device.
 */
TascamUs1641::TascamUs1641() {
	devs = NULL;
	dev = NULL;
	context = NULL;
	handle = NULL;
	returned = false;
	abort = false;

	// hardcode samplerate. The device supports 44100, 48000 and 96000
	// only 48000 has been tested
	samplerate = 48000;

	// the bulk transfer usb packets are 8192 bytes big.
	// just take something bigger, that fits
	buffersize = 16384;

	// create a usb control trasfer packet. This is used to send commands to the device.
	control_transfer = libusb_alloc_transfer(0);
	control_transfer_buffer = new unsigned char[buffersize];

	// big buffers to store audio data
	audio_out_buffer = new unsigned char[buffersize];
	audio_in_buffer = new unsigned char[buffersize];

	// initialize usb communication
	setup_usb();
	cout << "Setup complete ...\n";
	usleep(2000);
	// initialize device
	reset_usb();
	cout << "Reset complete ...\n";
}

/**
 * Free the usb handles and destroy all data structures.
 */
TascamUs1641::~TascamUs1641() {

	cout << "Destructing TascamUs1641.\n";
	abort = true;
	usleep(1000000);
	release_usb();
	libusb_free_transfer(control_transfer);
	delete[] control_transfer_buffer;
	delete[] audio_out_buffer;
	delete[] audio_in_buffer;
}

/**
 * Setup usb communication and request access to the device from the system.
 */
void TascamUs1641::setup_usb() {

	int error;
	error = libusb_init(&context);
	if (error < 0) {
		exit(error);
	}
	libusb_set_debug(context, 3);

	error = libusb_get_device_list(context, &devs);
	if (error < 0) {
		cout << "Device list empty!\n";
		exit(error);
	}

	handle = libusb_open_device_with_vid_pid(context, 0x644, 0x8017);
	if (handle == NULL) {
		cout << "Failed to get device handle of Tascam US-1641" << endl;
		exit(1);
	}
	dev = libusb_get_device(handle);

	error = libusb_kernel_driver_active(handle, 0);
	if (error) {
		cout << "Kernel driver active for interface 0: " << (int) error << endl;
		libusb_detach_kernel_driver(handle, 0);
		exit(1);
	}

	error = libusb_kernel_driver_active(handle, 1);
	if (error) {
		cout << "Kernel driver active for interface 1: " << (int) error << "\n";
		libusb_detach_kernel_driver(handle, 0);
		exit(1);
	}

	libusb_set_debug(context, 5);
}

/**
 * Release all usb data
 */
void TascamUs1641::release_usb() {
	libusb_release_interface(handle, 0);
	libusb_release_interface(handle, 1);
	libusb_free_device_list(devs, 1);
	libusb_unref_device(dev);
	libusb_close(handle);
	libusb_exit(context);
}

/**
 * Initialize device. This is also called, when the device lost synchronization due to
 * some system timeout.
 *
 * This function replays all events, that Wireshark logs when the original Tascam
 * driver connects to the device.
 * What these commands mean is mostly unknown to me.

 * The most interesting part from the users point of view is setting the samplerate.
 */
void TascamUs1641::reset_usb() {
	// reset usb communication
	libusb_reset_device(handle);
	libusb_set_configuration(handle, 1);
	libusb_ref_device(dev);
	libusb_claim_interface(handle, 0);
	libusb_claim_interface(handle, 1);
	libusb_set_interface_alt_setting(handle, 0, 1);
	libusb_set_interface_alt_setting(handle, 1, 1);

	usleep(2000);

	// send some "magic" commands to initialize the device

	// control transfer 1
	//bmRequestType = 0xc0 = USB_TYPE_VENDOR (0x40) | USB_DIR_IN (0x80)
	//				= 0xc0 = 11000000 = 192
	//bRequest 		= 0x56 = 01010110 = 86
	//wValue		= 0x00
	//wIndex 		= 0x00 => Endpoint 0
	//wLength		= 3
	submitAndWait_controlTransfer(0xc0, 0x56, 0x00, 0x00, 3);

	// control transfer 2	
	//bmRequestType = 0xc0 = USB_TYPE_VENDOR (0x40) | USB_DIR_IN (0x80)
	//				= 0xc0 = 11000000 = 192
	//bRequest 		= 0x56 = 01010110 = 86
	//wValue		= 0x00
	//wIndex 		= 0x00 => Endpoint 0
	//wLength		= 15
	submitAndWait_controlTransfer(0xc0, 0x56, 0x00, 0x00, 15);

	// control transfer 3	
	//bmRequestType = 0xc0 = USB_TYPE_VENDOR (0x40) | USB_DIR_IN (0x80)
	//				= 0xc0 = 11000000 = 192
	//bRequest 		= 0x49 = 01001001 = 73
	//wValue		= 0x00
	//wIndex 		= 0x00 => Endpoint 0
	//wLength		= 1
	submitAndWait_controlTransfer(0xc0, 0x49, 0x00, 0x00, 1);

	// control transfer 4	
	//bmRequestType = 0x40 = USB_TYPE_VENDOR (0x40) | USB_DIR_OUT (0x00)
	//				= 0x40 = 01000000 = 64
	//bRequest 		= 0x49 = 01001001 = 73
	//wValue		= 0x00
	//wIndex 		= 0x02 => Endpoint 2
	//wLength		= 0
	submitAndWait_controlTransfer(0x40, 0x49, 0x00, 0x02, 0);

	// control transfer 5	
	//bmRequestType = 0x22 = USB_TYPE_CLASS (0x20) | USB_RECIP_ENDPOINT (0x02) | USB_DIR_OUT (0x00)
	//				= 0x22 = 00100010 = 34
	//bRequest 		= 0x01 = UAC_SET_CUR
	//wValue		= 0x0100 = UAC_EP_CS_ATTR_SAMPLE_RATE
	//wIndex 		= 0x86 = Endpoint 6
	//wLength		= 3
	control_transfer_buffer[8 + 0] = samplerate;
	control_transfer_buffer[8 + 1] = samplerate >> 8;
	control_transfer_buffer[8 + 2] = samplerate >> 16;
	submitAndWait_controlTransfer(0x22, 0x01, 0x0100, 0x86, 3);

	// control transfer 6	
	//bmRequestType = 0x22 = USB_TYPE_CLASS (0x20) | USB_RECIP_ENDPOINT (0x02) | USB_DIR_OUT (0x00)
	//				= 0x22 = 00100010 = 34
	//bRequest 		= 0x01 = UAC_SET_CUR
	//wValue		= 0x0100 = UAC_EP_CS_ATTR_SAMPLE_RATE
	//wIndex 		= 0x02 = Endpoint 2
	//wLength		= 3
	control_transfer_buffer[8 + 0] = samplerate;
	control_transfer_buffer[8 + 1] = samplerate >> 8;
	control_transfer_buffer[8 + 2] = samplerate >> 16;
	submitAndWait_controlTransfer(0x22, 0x01, 0x0100, 0x02, 3);

	// control transfer 7
	//bmRequestType = 0xc0 = USB_TYPE_VENDOR (0x40) | USB_DIR_IN (0x80)
	//				= 0xc0 = 11000000 = 192
	//bRequest		= 0x41 = UAC__RES (0x40) | UAC_SET_CUR (0x01)
	//				= 0x41 = 01000001 = 65
	//wValue		= 0x0d00 = 00001101 00000000 = 3328
	//wIndex 		= 0x101 = 00000001 00000001 = 257
	//wLength		= 5
	submitAndWait_controlTransfer(0xc0, 0x41, 0x0d00, 257, 5);

	// control transfer 8	
	//bmRequestType = 0x22 = USB_TYPE_CLASS (0x20) | USB_RECIP_ENDPOINT (0x02) | USB_DIR_OUT (0x00)
	//				= 0x22 = 00100010 = 34
	//bRequest 		= 0x01 = UAC_SET_CUR
	//wValue		= 0x0100 = UAC_EP_CS_ATTR_SAMPLE_RATE
	//wIndex 		= 0x86 = Endpoint 6
	//wLength		= 3
	control_transfer_buffer[8 + 0] = samplerate;
	control_transfer_buffer[8 + 1] = samplerate >> 8;
	control_transfer_buffer[8 + 2] = samplerate >> 16;
	submitAndWait_controlTransfer(0x22, 0x01, 0x0100, 0x86, 3);

	// control transfer 9
	//bmRequestType = 0xa2 = USB_TYPE_CLASS (0x20) | USB_RECIP_ENDPOINT (0x02) | USB_DIR_IN (0x80)
	//				= 0xa2 = 10100010 = 162
	//bRequest 		= 0x81 = UAC_GET_CUR
	//wValue		= 0x0100 = UAC_EP_CS_ATTR_SAMPLE_RATE
	//wIndex 		= 0x86 = Endpoint 6
	//wLength		= 3
	submitAndWait_controlTransfer(0xa2, 0x81, 0x0100, 0x86, 3);

	// control transfer 10
	//bmRequestType = 0xc0 = USB_TYPE_VENDOR (0x40) | USB_DIR_IN (0x80)
	//				= 0xc0 = 11000000 = 192
	//bRequest		= 0x49 = 01001001 = 73
	//wValue		= 0x0000
	//wIndex 		= 0x00 => Endpoint 0
	//wLength		= 1
	submitAndWait_controlTransfer(0xc0, 0x49, 0x0000, 0, 1);

	// control transfer 11	
	//bmRequestType = 0x40 = USB_TYPE_VENDOR (0x40) | USB_DIR_OUT (0x00)
	//				= 0x40 = 01000000 = 64
	//bRequest 		= 0x49 = 01001001 = 73
	//wValue		= 0x0032 = 00000000 00110010 = 50
	//wIndex 		= 0x00 => Endpoint 0
	//wLength		= 0
	submitAndWait_controlTransfer(0x40, 0x49, 0x0032, 0, 0);

	// control transfer 12
	//bmRequestType = 0xc0 = USB_TYPE_VENDOR (0x40) | USB_DIR_IN (0x80)
	//				= 0xc0 = 11000000 = 192
	//bRequest 		= 0x49 = 01001001 = 73
	//wValue		= 0x0032 = 00000000 00110010 = 50
	//wIndex 		= 0x00 => Endpoint 0
	//wLength		= 1
	submitAndWait_controlTransfer(0xc0, 0x49, 0x0032, 0, 1);

	// control transfer 13
	//bmRequestType = 0x02 = USB_RECIP_ENDPOINT (0x02)
	//				= 0x02 = 00000010 = 2
	//bRequest 		= 0x01 = 10000000 = 1
	//wValue		= 0x0000
	//wIndex 		= 0x00 => Endpoint 6
	//wLength		= 0
	submitAndWait_controlTransfer(0x02, 0x01, 0x0000, 0x86, 0);

}

/**
 * When the usb transfers timeout or the device doesn't get audio packets fast enough,
 * it has to be resumed.
 */
void TascamUs1641::resumePlayback() {
	cerr << "playback resumed" << endl << flush;
	submit_controlTransfer(0x40, 0x49, 0x0032, 0, 0);
}

/**
 * Submit a control transfer to the usb system and wait until this packet has
 * been processed.
 */
void TascamUs1641::submitAndWait_controlTransfer(uint8_t bmRequestType,
		uint8_t bRequest, uint16_t wValue, uint16_t wIndex, uint16_t wLength) {
	submit_controlTransfer(bmRequestType, bRequest, wValue, wIndex, wLength);
	controlTransferReturned = false;
	while (!controlTransferReturned)
		libusb_handle_events(context);
}

/**
 * Submit a control transfer to the usb system and continue with the program flow, i.e.
 * do not wait until the packet has been processed.
 */
void TascamUs1641::submit_controlTransfer(uint8_t bmRequestType,
		uint8_t bRequest, uint16_t wValue, uint16_t wIndex, uint16_t wLength) {
	libusb_fill_control_setup(control_transfer_buffer, bmRequestType, bRequest,
			wValue, wIndex, wLength);
	libusb_fill_control_transfer(control_transfer, handle,
			control_transfer_buffer, callback_controlTransfer, NULL, 1000);
	control_transfer->user_data = this;
	libusb_submit_transfer(control_transfer);
}

/**
 * This method is called when a control transfer has been completed.
 */
void TascamUs1641::callback_controlTransfer(struct libusb_transfer * transfer) {
	TascamUs1641 * tascam = (TascamUs1641*) transfer->user_data;
	tascam->controlTransferReturned = true;
}

/**
 * Prepare a data packet, that sends audio to the device using an isochronous transfer.
 */
struct libusb_transfer * TascamUs1641::allocate_output_transfer() {
	struct libusb_transfer * transfer = libusb_alloc_transfer(8);

	if (!transfer)
		exit(1);

	// prepare the data packet and populate the neccessary fields
	transfer->dev_handle = this->handle;
	transfer->endpoint = 0x02;
	transfer->type = LIBUSB_TRANSFER_TYPE_ISOCHRONOUS;
	transfer->timeout = 100;
	transfer->buffer = audio_out_buffer;
	memset(transfer->buffer, 0, buffersize);
	transfer->user_data = this;
	transfer->length = 1024;
	transfer->callback = TascamUs1641::callback_audioout;
	transfer->num_iso_packets = 8;

	for (unsigned int i = 0; i < 8; i++) {
		struct libusb_iso_packet_descriptor *desc =
				&transfer->iso_packet_desc[i];
		desc->length = 72;
	}

	return transfer;
}

/**
 * Prepare a data packet, that reads data from the device using a bulk transfer.
 */
struct libusb_transfer * TascamUs1641::allocate_input_transfer() {
	struct libusb_transfer * transfer = libusb_alloc_transfer(1);

	if (!transfer)
		exit(1);

	// prepare the data packet and populate the neccessary fields
	transfer->dev_handle = this->handle;
	transfer->endpoint = 0x86;
	transfer->type = LIBUSB_TRANSFER_TYPE_BULK;
	transfer->timeout = 20;
	transfer->buffer = audio_in_buffer;
	memset(transfer->buffer, 0, buffersize);
	transfer->user_data = this;
	transfer->length = 8192;
	transfer->callback = TascamUs1641::callback_audioin;

	return transfer;
}

/**
 * This method is called after an audio packet has been transferred to the device.
 * It clears the data buffer and sends a new audio packet.
 */
void TascamUs1641::callback_audioout(struct libusb_transfer * transfer) {

	TascamUs1641 * tascam = (TascamUs1641*) transfer->user_data;

	// clear data buffer
	transfer->buffer = tascam->audio_out_buffer;
	memset(transfer->buffer, 0, tascam->buffersize);

	// send audio packet containing silence.
	if (!tascam->abort)
		libusb_submit_transfer(transfer);
}

/**
 * This method is called after an audio packet has been transferred from the device.
 * If the transfer fails, the device is reset. Request the next data packet afterwards.
 */
void TascamUs1641::callback_audioin(struct libusb_transfer * transfer) {

	TascamUs1641 * tascam = (TascamUs1641*) transfer->user_data;

	// if the usb connection breaks down, resume it
	if (transfer->status != 0) {
		tascam->resumePlayback();
	}

	transfer->buffer = tascam->audio_in_buffer;

	// Request next data packet.
	if (!tascam->abort)
		libusb_submit_transfer(transfer);
}

/**
 * "Main loop" of the device driver. It starts the transfers and goes into an
 * infinite loop that polls the usb system for events.
 */
void TascamUs1641::play() {
	transfer1 = allocate_output_transfer();
	transfer2 = allocate_output_transfer();
	transfer3 = allocate_input_transfer();
	if (libusb_submit_transfer(transfer1) < 0 || libusb_submit_transfer(
			transfer2) < 0 || libusb_submit_transfer(transfer3) < 0)
		printf("Submitting play transfers failed.");

	while (!abort) {
		if (libusb_handle_events(context) < 0)
			break;
	}
}
